'''
Load data
'''

#%% Housekeeping
import numpy as np
import sciris as sc
from . import country_age_data    as cad
from . import state_age_data      as sad
from . import household_size_data as hsd

__all__ = ['get_country_aliases', 'map_entries', 'get_age_distribution', 'get_household_size']


def get_country_aliases():
    ''' Define aliases for countries with odd names in the data '''
    country_mappings = {
       'Bolivia':        'Bolivia (Plurinational State of)',
       'Burkina':        'Burkina Faso',
       'Cape Verde':     'Cabo Verdeo',
       'Hong Kong':      'China, Hong Kong Special Administrative Region',
       'Macao':          'China, Macao Special Administrative Region',
       "Cote d'Ivore":   'Côte d’Ivoire',
       "Ivory Coast":    'Côte d’Ivoire',
       'DRC':            'Democratic Republic of the Congo',
       'Iran':           'Iran (Islamic Republic of)',
       'Laos':           "Lao People's Democratic Republic",
       'Micronesia':     'Micronesia (Federated States of)',
       'Korea':          'Republic of Korea',
       'South Korea':    'Republic of Korea',
       'Moldova':        'Republic of Moldova',
       'Russia':         'Russian Federation',
       'Palestine':      'State of Palestine',
       'Syria':          'Syrian Arab Republic',
       'Taiwan':         'Taiwan Province of China',
       'Macedonia':      'The former Yugoslav Republic of Macedonia',
       'UK':             'United Kingdom of Great Britain and Northern Ireland',
       'United Kingdom': 'United Kingdom of Great Britain and Northern Ireland',
       'Tanzania':       'United Republic of Tanzania',
       'USA':            'United States of America',
       'United States':  'United States of America',
       'Venezuela':      'Venezuela (Bolivarian Republic of)',
       'Vietnam':        'Viet Nam',
        }

    return country_mappings # Convert to lowercase


def map_entries(json, location):
    '''
    Find a match between the JSON file and the provided location(s).

    Args:
        json (list or dict): the data being loaded
        location (list or str): the list of locations to pull from
    '''

    # The data have slightly different formats: list of dicts or just a dict
    countries = [key.lower() for key in json.keys()]

    # Set parameters
    if location is None:
        location = countries
    else:
        location = sc.promotetolist(location)

    # Define a mapping for common mistakes
    mapping = get_country_aliases()
    mapping = {key.lower(): val.lower() for key, val in mapping.items()}

    entries = {}
    for loc in location:
        lloc = loc.lower()
        if lloc not in countries and lloc in mapping:
            lloc = mapping[lloc]
        try:
            ind = countries.index(lloc)
            entry = list(json.values())[ind]
            entries[loc] = entry
        except ValueError as E:
            suggestions = sc.suggest(loc, countries, n=4)
            if suggestions:
                errormsg = f'Location "{loc}" not recognized, did you mean {suggestions}? ({str(E)})'
            else:
                errormsg = f'Location "{loc}" not recognized ({str(E)})'
            raise ValueError(errormsg)

    return entries


def get_age_distribution(location=None):
    '''
    Load age distribution for a given country or countries.

    Args:
        location (str or list): name of the country or countries to load the age distribution for

    Returns:
        age_data (array): Numpy array of age distributions, or dict if multiple locations
    '''

    # Load the raw data
    country_json = cad.get()
    state_json   = sad.get()
    json = {**state_json, **country_json}
    entries = map_entries(json, location)

    max_age = 99
    result = {}
    for loc,age_distribution in entries.items():
        total_pop = sum(list(age_distribution.values()))
        local_pop = []

        for age, age_pop in age_distribution.items():
            if age[-1] == '+':
                val = [int(age[:-1]), max_age, age_pop/total_pop]
            else:
                ages = age.split('-')
                val = [int(ages[0]), int(ages[1]), age_pop/total_pop]
            local_pop.append(val)
        result[loc] = np.array(local_pop)

    if len(result) == 1:
        result = list(result.values())[0]

    return result


def get_household_size(location=None):
    '''
    Load household size distribution for a given country or countries.

    Args:
        location (str or list): name of the country or countries to load the age distribution for

    Returns:
        house_size (float): Size of household, or dict if multiple locations
    '''
    # Load the raw data
    json = hsd.get()

    result = map_entries(json, location)
    if len(result) == 1:
        result = list(result.values())[0]

    return result



"""
Tests of simulation parameters from
../../covasim/README.md
"""
import unittest

from unittest_support_classes import CovaSimTest, TestProperties

ResKeys = TestProperties.ResultsDataKeys
ParamKeys = TestProperties.ParameterKeys


class DiseaseProgressionTests(CovaSimTest):
    def setUp(self):
        super().setUp()
        pass

    def tearDown(self):
        super().tearDown()
        pass

    @unittest.skip("Not fixing deviation tests now")
    def test_exposure_to_infectiousness_delay_deviation_scaling(self):
        """
        Configure exposure to infectiousness delay to 1/2 sim
        length, and std_dev to 0. Verify that every pop_infected
        at start goes infectious on the same day
        """
        total_agents = 500
        self.set_everyone_infected(total_agents)
        sim_dur = 60
        exposed_to_infectious_delay = 30
        std_devs = [0, .5, 1, 2, 4]  # Keep values in order

        prev_first_day = None
        prev_peak_val = None
        prev_last_day = None
        prev_stddev = None

        for TEST_stddev in std_devs:

            serial_dev_zero = {
                TestProperties.ParameterKeys.SimulationKeys.number_simulated_days: sim_dur,
                TestProperties.ParameterKeys.ProgressionKeys.exposed_to_infectious: exposed_to_infectious_delay,
                TestProperties.ParameterKeys.ProgressionKeys.exposed_to_infectious_std: TEST_stddev,
                TestProperties.ParameterKeys.MortalityKeys.use_cfr_by_age: False,
                TestProperties.ParameterKeys.MortalityKeys.prob_infected_symptomatic: 0
            }
            self.run_sim(serial_dev_zero)
            infectious_channel = self.get_full_result_channel(
                ResKeys.infectious_at_timestep
            )
            if TEST_stddev == 0:
                for t in range(0, exposed_to_infectious_delay):
                    today_infectious = infectious_channel[t]
                    self.assertEqual(today_infectious, 0,
                                     msg=f"With std_dev 0, there should be no infectious"
                                         f"prior to {exposed_to_infectious_delay} + 1. At {t} we had"
                                         f" {today_infectious}.")
                    pass
                prev_infectious = infectious_channel[exposed_to_infectious_delay]
                self.assertEqual(prev_infectious, total_agents,
                                 msg=f"At {exposed_to_infectious_delay} + 1, should have {total_agents}"
                                     f" infectious individuals with std_dev 0. Got {prev_infectious}.")
                prev_first_day = exposed_to_infectious_delay  # were zero infectious before this
                prev_peak_val = total_agents  # this is everyone
                prev_last_day = exposed_to_infectious_delay  # and there is nobody left
                for t in range(exposed_to_infectious_delay + 1, len(infectious_channel)):
                    today_infectious = infectious_channel[t]
                    self.assertLessEqual(today_infectious, prev_infectious,
                                         msg="With std_dev 0, after initial infectious conversion,"
                                             " infectiousness should only decline. Not the case"
                                             f" at t: {t}")
                    pass
                prev_stddev = TEST_stddev
            else:
                curr_first_day = None
                curr_peak_day = None
                curr_peak_val = 0
                curr_last_day = None
                for t in range(len(infectious_channel)):
                    today_infectious = infectious_channel[t]
                    if today_infectious > 0:
                        if not curr_first_day:
                            curr_first_day = t
                            curr_peak_day = t
                            curr_peak_val = today_infectious
                        elif today_infectious > curr_peak_val:
                            curr_peak_day = t
                            curr_peak_val = today_infectious
                        else:
                            curr_last_day = t
                            pass
                        pass
                    pass
                if self.is_debugging:
                    print(f"TEST_stddev: {TEST_stddev}")
                    print(f"current first day: {curr_first_day}")
                    print(f"current peak day: {curr_peak_day}, curr_peak_val: {curr_peak_val}")
                    print(f"current last day: {curr_last_day}")
                    print(f"infectious_channel: {infectious_channel}")
                self.assertLessEqual(curr_first_day, prev_first_day,
                                msg=f"With stddev {TEST_stddev}, first infectious day {curr_first_day}"
                                    f" should be lower than previous {prev_first_day}"
                                    f" that came with stddev {prev_stddev}.")
                self.assertLess(curr_peak_val, prev_peak_val,
                                 msg=f"With stddev {TEST_stddev}, the peak value {curr_peak_val}"
                                     f" from peak day {curr_peak_day} should be less than the previous"
                                     f" peak value {prev_peak_val} from stddev {prev_stddev}.")
                self.assertGreaterEqual(curr_last_day, prev_last_day,
                                   msg=f"With stddev {TEST_stddev}, last infectious conversion {curr_last_day}"
                                       f" should be greater than previous {prev_last_day}"
                                       f" that came with stddev {prev_stddev}.")
                # Having passed all assertions, reset expectations
                prev_first_day = curr_first_day
                prev_peak_val = curr_peak_val
                prev_last_day = curr_last_day
        pass

    def test_exposure_to_infectiousness_delay_scaling(self):
        """
        Set exposure to infectiousness early simulation, mid simulation,
        late simulation. Set std_dev to zero. Verify move to infectiousness
        moves later as delay is longer.
        Depends on delay deviation test
        """
        total_agents = 500
        self.set_everyone_infected(total_agents)
        sim_dur = 60
        exposed_delays = [1, 2, 5, 15, 20, 25, 30]  # Keep values in order
        std_dev = 0
        for exposed_delay in exposed_delays:
            self.set_duration_distribution_parameters(
                duration_in_question=ParamKeys.ProgressionKeys.DurationKeys.exposed_to_infectious,
                par1=exposed_delay,
                par2=std_dev
            )
            prob_dict = {
                TestProperties.ParameterKeys.ProgressionKeys.ProbabilityKeys.RelativeProbKeys.inf_to_symptomatic_probability: 0
            }
            self.set_simulation_prognosis_probability(prob_dict)
            serial_delay = {
                TestProperties.ParameterKeys.SimulationKeys.number_simulated_days: sim_dur
            }
            self.run_sim(serial_delay)
            infectious_channel = self.get_full_result_channel(
                ResKeys.infectious_at_timestep
            )
            agents_on_infectious_day = infectious_channel[exposed_delay]
            if self.is_debugging:
                print(f"Delay: {exposed_delay}")
                print(f"Agents turned: {agents_on_infectious_day}")
                print(f"Infectious channel {infectious_channel}")
                pass
            for t in range(len(infectious_channel)):
                current_infectious = infectious_channel[t]
                if t < exposed_delay:
                    self.assertEqual(current_infectious, 0,
                                     msg=f"All {total_agents} should turn infectious at t: {exposed_delay}"
                                         f" instead got {current_infectious} at t: {t}")
                elif t == exposed_delay:
                    self.assertEqual(infectious_channel[exposed_delay], total_agents,
                                     msg=f"With stddev 0, all {total_agents} agents should turn infectious "
                                         f"on day {exposed_delay}, instead got {agents_on_infectious_day}. ")
        pass

    @unittest.skip("Not fixing deviation tests now")
    def test_infection_duration_deviation_scaling(self):
        """
        Like deviation zero test, but with expected duration midsim.
        Verification like the exposure_to_infectiouness_delay
        deviation scaling test (as std_dev goes up, first conversion
        earlier and last conversion later)
        """
        total_agents = 500
        exposed_delay = 1
        self.set_everyone_infectious_same_day(num_agents=total_agents,
                                              days_to_infectious=exposed_delay)
        infectious_duration = 30
        infectious_duration_stddevs = [0, 1, 2, 4]  # Keep values in order
        for TEST_std in infectious_duration_stddevs:
            test_config = {
                TestProperties.ParameterKeys.ProgressionKeys.infectiousness_duration: infectious_duration,
                TestProperties.ParameterKeys.ProgressionKeys.infectiousness_duration_std: TEST_std
            }
            self.run_sim(params_dict=test_config)
            recoveries_channel = self.get_full_result_channel(
                TestProperties.ResultsDataKeys.recovered_at_timestep_WHAT
            )
            if TEST_std == 0:
                # do the zero stuff
                everyone_recovers_day = infectious_duration + exposed_delay
                for t in range(exposed_delay, everyone_recovers_day):
                    today_recoveries = recoveries_channel[t]
                    self.assertEqual(today_recoveries, 0,
                                     msg=f"With std_dev 0, there should be no infectious"
                                         f"prior to {everyone_recovers_day}. At {t} we had"
                                         f" {today_recoveries}.")
                    pass
                prev_recoveries = recoveries_channel[infectious_duration + exposed_delay]
                self.assertEqual(prev_recoveries, total_agents,
                                 msg=f"At {everyone_recovers_day}, should have {total_agents}"
                                     f" recovered individuals with std_dev 0. Got {prev_recoveries}.")
                prev_first_day = everyone_recovers_day # were zero before this
                prev_peak_val = total_agents
                prev_last_day = everyone_recovers_day # and there is nobody left
                for t in range(everyone_recovers_day + 1, len(recoveries_channel)):
                    today_recoveries = recoveries_channel[t]
                    self.assertEqual(today_recoveries, 0,
                                     msg="With std_dev 0, after initial recovery day,"
                                         " recoveries should remain 0. Not the case"
                                         f" at t: {t} with recoveries: {today_recoveries}")
                    pass
                prev_stddev = TEST_std
                pass
            else:
                curr_first_day = None
                curr_peak_day = None
                curr_peak_val = 0
                curr_last_day = 0
                for t in range(infectious_duration, len(recoveries_channel)):
                    today_recoveries = recoveries_channel[t]
                    if today_recoveries > 0:
                        if not curr_first_day:
                            curr_first_day = t
                        elif today_recoveries > curr_peak_val:
                            curr_peak_day = t
                            curr_peak_val = today_recoveries
                        else:
                            curr_last_day = t
                            pass
                        pass
                    pass

                if self.is_debugging:
                    print(f"TEST_stddev: {TEST_std}")
                    print(f"current first day: {curr_first_day}")
                    print(f"current peak day: {curr_peak_day}, curr_peak_val: {curr_peak_val}")
                    print(f"current last day: {curr_last_day}")
                    print(f"recoveries_channel: {recoveries_channel}")
                self.assertLessEqual(curr_first_day, prev_first_day,
                                msg=f"With stddev {TEST_std}, first infectious day {curr_first_day}"
                                    f" should be lower than previous {prev_first_day}"
                                    f" that came with stddev {prev_stddev}.")
                self.assertLess(curr_peak_val, prev_peak_val,
                                 msg=f"With stddev {TEST_std}, the peak value {curr_peak_val}"
                                     f"from peak day {curr_peak_day} should be less than the previous"
                                     f"peak value {prev_peak_val} from stddev {prev_stddev}.")
                self.assertGreaterEqual(curr_last_day, prev_last_day,
                                   msg=f"With stddev {TEST_std}, last infectious conversion {curr_last_day}"
                                       f" should be greater than previous {prev_last_day}"
                                       f" that came with stddev {prev_stddev}.")
                pass
            pass
        pass

    def test_mild_infection_duration_scaling(self):
        """
        Make sure that all initial infected cease being infected
        on following day. Std_dev 0 will help here
        """
        total_agents = 500
        exposed_delay = 1
        self.set_everyone_infectious_same_day(num_agents=total_agents,
                                              days_to_infectious=exposed_delay)
        prob_dict = {
            ParamKeys.ProgressionKeys.ProbabilityKeys.RelativeProbKeys.inf_to_symptomatic_probability: 0.0
        }
        self.set_simulation_prognosis_probability(prob_dict)
        infectious_durations = [1, 2, 5, 10, 20] # Keep values in order
        infectious_duration_stddev = 0
        for TEST_dur in infectious_durations:
            recovery_day = exposed_delay + TEST_dur
            self.set_duration_distribution_parameters(
                duration_in_question=ParamKeys.ProgressionKeys.DurationKeys.infectious_asymptomatic_to_recovered,
                par1=TEST_dur,
                par2=infectious_duration_stddev
            )
            self.run_sim()
            recoveries_channel = self.get_full_result_channel(
                TestProperties.ResultsDataKeys.recovered_at_timestep
            )
            recoveries_on_recovery_day = recoveries_channel[recovery_day]
            if self.is_debugging:
                print(f"Delay: {recovery_day}")
                print(f"Agents turned: {recoveries_on_recovery_day}")
                print(f"Recoveries channel {recoveries_channel}")
            self.assertEqual(recoveries_channel[recovery_day], total_agents,
                             msg=f"With stddev 0, all {total_agents} agents should turn infectious "
                                 f"on day {recovery_day}, instead got {recoveries_on_recovery_day}. ")

        pass

    @unittest.skip("Not fixing deviation tests now")
    def test_time_to_die_deviation_scaling(self):
        total_agents = 500
        self.set_everyone_is_going_to_die(num_agents=total_agents)

        time_to_die_duration = 30
        time_to_die_duration_stddevs = [0, 1, 2, 4]

        prev_first_death_day = None
        prev_peak_death_count = 0
        prev_last_death_day = None

        for TEST_std in time_to_die_duration_stddevs:
            test_config = {
                TestProperties.ParameterKeys.MortalityKeys.time_to_death: time_to_die_duration,
                TestProperties.ParameterKeys.MortalityKeys.time_to_death_std: TEST_std
            }
            self.run_sim(params_dict=test_config)
            deaths_today_channel = self.get_full_result_channel(
                TestProperties.ResultsDataKeys.deaths_daily
            )
            death_timer_starts = 0
            if TEST_std == 0:
                expected_death_day = death_timer_starts + time_to_die_duration
                for t in range(death_timer_starts, expected_death_day):
                    curr_deaths = deaths_today_channel[t]
                    self.assertEqual(curr_deaths, 0,
                                     msg=f"With std 0, expected no deaths until {expected_death_day},"
                                         f" but saw {curr_deaths} at time {t}.")
                    pass
                curr_deaths = deaths_today_channel[expected_death_day]
                self.assertEqual(curr_deaths, total_agents,
                                 msg=f"With std 0, expected {total_agents} deaths on"
                                     f" day {t}  but saw {curr_deaths} instead.")
                prev_first_death_day = expected_death_day # Last is curr day if rest were zero
                prev_peak_death_count = curr_deaths # this is everybody
                for t in range(expected_death_day + 1, len(deaths_today_channel)):
                    curr_deaths = deaths_today_channel[t]
                    self.assertEqual(curr_deaths, 0,
                                     msg=f"With std 0, expected no deaths after {expected_death_day},"
                                         f" but saw {curr_deaths} at time {t}.")
                    pass
                prev_last_death_day = prev_first_death_day # Last same as first if rest are 0
                pass
            else:
                curr_first_death_day = None
                curr_peak_death_count = 0
                curr_last_death_day = None

                for t in range(death_timer_starts, len(deaths_today_channel)):
                    today_deaths = deaths_today_channel[t]
                    if today_deaths > 0:
                        if not curr_first_death_day:
                            curr_first_death_day = t
                        elif today_deaths > curr_peak_death_count:
                            curr_peak_death_count = today_deaths
                        else:
                            curr_last_death_day = t
                            pass
                        pass
                    pass
                self.assertLessEqual(curr_first_death_day, prev_first_death_day)
                self.assertLess(curr_peak_death_count, prev_peak_death_count)
                self.assertGreaterEqual(curr_last_death_day, prev_last_death_day)
                prev_first_death_day = curr_first_death_day
                prev_peak_death_count = curr_peak_death_count
                prev_last_death_day = curr_last_death_day
                pass
        pass

    def test_time_to_die_duration_scaling(self):
        total_agents = 500
        self.set_everyone_critical(num_agents=500, constant_delay=0)
        prob_dict = {
            ParamKeys.ProgressionKeys.ProbabilityKeys.RelativeProbKeys.crt_to_death_probability: 1.0
        }
        self.set_simulation_prognosis_probability(prob_dict)

        time_to_die_durations = [1, 2, 5, 10, 20]
        time_to_die_stddev = 0

        for TEST_dur in time_to_die_durations:
            self.set_duration_distribution_parameters(
                duration_in_question=ParamKeys.ProgressionKeys.DurationKeys.critical_to_death,
                par1=TEST_dur,
                par2=time_to_die_stddev
            )
            self.run_sim()
            deaths_today_channel = self.get_full_result_channel(
                TestProperties.ResultsDataKeys.deaths_daily
            )
            for t in range(len(deaths_today_channel)):
                curr_deaths = deaths_today_channel[t]
                if t < TEST_dur:
                    self.assertEqual(curr_deaths, 0,
                                     msg=f"With std 0, all {total_agents} agents should die on "
                                         f"t: {TEST_dur}. Got {curr_deaths} at t: {t}")
                elif t == TEST_dur:
                    self.assertEqual(curr_deaths, total_agents,
                                     msg=f"With std 0, all {total_agents} agents should die at t:"
                                         f" {TEST_dur}, got {curr_deaths} instead.")
                else:
                    self.assertEqual(curr_deaths, 0,
                                     msg=f"With std 0, all {total_agents} agents should die at t:"
                                         f" {TEST_dur}, got {curr_deaths} at t: {t}")
        pass

    @unittest.skip("NYI")
    def test_exposed_to_symptomatic_delay_deviation_scaling(self):
        total_agents = 500
        self.set_everyone_symptomatic(num_agents=total_agents)

        incubation_duration = 30
        incubation_duration_stds = [0, 1, 2, 4]
        for TEST_std in incubation_duration_stds:
            test_config = {
                TestProperties.ParameterKeys.ProgressionKeys.exposed_to_symptomatic: incubation_duration,
                TestProperties.ParameterKeys.ProgressionKeys.exposed_to_symptomatic_std: TEST_std
            }
            self.run_sim(params_dict=test_config)
            symptomatic_count_channel = self.get_full_result_channel(
                TestProperties.ResultsDataKeys.symptomatic_at_timestep
            )
            if TEST_std == 0:
                for t in range(1, len(symptomatic_count_channel)):
                    # get the people sympto at current ts, subtract from previous
                    # if that number is zero...
                    # proceed as normal / above
                    pass
            else:
                pass
            pass
        pass
    pass

if __name__ == '__main__':
    unittest.main()


import json
from covasim import Sim, parameters, test_prob, contact_tracing, sequence

def test_basic_contact_trace():

    tweaking = {
        'n_days': 60,
        'pop_size': 2000,
        'pop_type': 'hybrid'
    }

    start_days = []
    interventions = []

    test_prob_start_day = 35
    prob_test_perfect_symptomatic = test_prob(symp_prob=1.0, symp_quar_prob=1.0, asymp_quar_prob=1.0,
                                              test_sensitivity=1.0, test_delay=1, start_day=test_prob_start_day)
    start_days.append(test_prob_start_day)
    interventions.append(prob_test_perfect_symptomatic)

    contact_trace_start_day = 35
    all_contacts = {
        'h': 1.0,
        'w': 1.0,
        's': 1.0,
        'c': 1.0
    }
    single_day_delays = {
        'h': 1,
        'w': 1,
        's': 1,
        'c': 1
    }
    brutal_contact_trace = contact_tracing(trace_probs=all_contacts,
                                           trace_time=single_day_delays,
                                           start_day=contact_trace_start_day)
    start_days.append(contact_trace_start_day)
    interventions.append(brutal_contact_trace)

    tweaking['interventions'] = interventions

    test_stuff_simulation = Sim(pars=tweaking)
    test_stuff_simulation.run()

    test_stuff_results = test_stuff_simulation.to_json(tostring=False)
    with open("DEBUG_test_intervention_list_simulation.json","w") as outfile:
        json.dump(test_stuff_results, outfile, indent=4, sort_keys=True)
        pass

    assert test_stuff_results["results"]["n_symptomatic"][test_prob_start_day] > 0  # If there are symptomatics
    assert sum(test_stuff_results["results"]["new_tests"][test_prob_start_day:test_prob_start_day + 5]) > 0  # then there should be tests
    assert sum(test_stuff_results["results"]["new_diagnoses"][test_prob_start_day + 1:test_prob_start_day + 6]) > 0 # and some diagnoses
    assert sum(test_stuff_results["results"]["new_diagnoses"][test_prob_start_day + 2:test_prob_start_day + 7]) > 0 # and therefore some quarantined
    pass


if __name__ == "__main__":
    test_basic_contact_trace()
    pass




from cova_epi_scraper import Scraper
import pandas as pd


class ECDPScraper(Scraper):
    def create_date(self):
        self.df["date"] = pd.to_datetime(self.df[["year", "month", "day"]])


pars = dict()
pars['title'] = 'European Centre for Disease Prevention and Control Covid-19 Data Scraper'
pars['load_path'] = 'https://opendata.ecdc.europa.eu/covid19/casedistribution/csv'

pars['output_folder'] = 'epi_data/ecdp'

pars['renames'] = dict()
pars['renames']['countriesAndTerritories'] = 'key'
pars['renames']['cases']                   = 'new_diagnoses'
pars['renames']['deaths']                  = 'new_deaths'
pars['renames']['popData2018']             = 'population'

pars['fields_to_drop'] = [
    'dateRep',
    'month',
    'year',
    'geoId',
    'countryterritoryCode'
]

ECDPScraper(pars).scrape()




import covasim as cv
import numpy as np


s1s = []
s2s = []
s3s = []
ps = [5000, 10000, 20000, 50000, 100000][1]
pi = 100
nd = 40
for i in [1, 2, 3]:
    print(i)
    s1 = cv.Sim(pop_size=ps, n_days=nd, pop_infected=pi, pop_type='random', rand_seed=i, verbose=0)
    s1.run()
    s1s.append(s1.summary['cum_infections'])
    s2 = cv.Sim(pop_size=ps, n_days=nd, pop_infected=pi, pop_type='hybrid', rand_seed=i, verbose=0)
    s2.run()
    s2s.append(s2.summary['cum_infections'])
    s3 = cv.Sim(pop_size=ps, n_days=nd, pop_infected=pi, pop_type='synthpops', rand_seed=i, verbose=0)
    s3.run()
    s3s.append(s3.summary['cum_infections'])

print(s1s)
print(s2s)
print(s3s)
print(np.mean(s1s))
print(np.mean(s2s))
print(np.mean(s3s))

s2.people.make_detailed_transtree()
s3.people.make_detailed_transtree()
tt_hy = s2.people.transtree.detailed
tt_sp = s3.people.transtree.detailed

for i,tt in enumerate([tt_hy, tt_sp]):
    print(['hy', 'sp'][i])
    for lk in 'hswc':
        infs = [d for d in tt if d and d['layer']==lk]
        print(lk, len(infs))

s2.people.transtree.plot()
s3.people.transtree.plot()


"""
Tests of simulation parameters from
../../covasim/README.md
"""
import unittest

from unittest_support_classes import CovaSimTest, TestProperties

class DiagnosticTestingTests(CovaSimTest):
    def setUp(self):
        super().setUp()
        pass

    def tearDown(self):
        super().tearDown()
        pass

    @unittest.skip("NYI")
    def test_number_of_daily_tests(self):
        """
        Set sensitivity to 1 and number_daily_tests
        to number_of_agents. Diagnoses should == infections
        """
        pass

    @unittest.skip("NYI")
    def test_daily_test_sensitivity(self):
        """
        With sensitivity to 0, should see no diagnoses
        With sensitivity 0.3 should see more
        With sensitivity 0.7 should see more
        Sensitivity 1 is covered above
        """
        pass

    @unittest.skip("NYI")
    def test_symptomatic_testing_multiplier(self):
        """
        Set whole population to symptomatic, and
        Sensitivity to low positive (0.25)
        See this scale up
        0 no diagnoses
        1.0 more (more than zero)
        2.0 about half (more than 1)
        4.0 about all (more than 2)
        """
        pass

    @unittest.skip("NYI")
    def test_trace_testing_multiplier(self):
        """
        Set number of contacts to a high value
        To see a stronger effect (probably)
        """
        pass


"""
Tests of simulation parameters from
../../covasim/README.md
"""
import unittest
import pytest

from unittest_support_classes import CovaSimTest, TestProperties

DProgKeys = TestProperties.ParameterKeys.ProgressionKeys
TransKeys = TestProperties.ParameterKeys.TransmissionKeys
TSimKeys = TestProperties.ParameterKeys.SimulationKeys
ResKeys = TestProperties.ResultsDataKeys


class DiseaseMortalityTests(CovaSimTest):
    def setUp(self):
        super().setUp()
        pass

    def tearDown(self):
        super().tearDown()
        pass

    def test_default_death_prob_one(self):
        """
        Infect lots of people with cfr one and short time to die
        duration. Verify that everyone dies, no recoveries.
        """
        total_agents = 500
        self.set_everyone_is_going_to_die(num_agents=total_agents)
        self.run_sim()
        recoveries_at_timestep_channel = self.get_full_result_channel(
            ResKeys.recovered_at_timestep
        )
        recoveries_cumulative_channel = self.get_full_result_channel(
            ResKeys.recovered_cumulative
        )
        recovery_channels = [
            recoveries_at_timestep_channel,
            recoveries_cumulative_channel
        ]
        for c in recovery_channels:
            for t in range(len(c)):
                self.assertEqual(0, c[t],
                                 msg=f"There should be no recoveries"
                                     f" with death_prob 1.0. Channel {c} had "
                                     f" bad data at t: {t}")
                pass
            pass
        cumulative_deaths = self.get_day_final_channel_value(
            ResKeys.deaths_cumulative
        )
        self.assertEqual(cumulative_deaths, total_agents,
                         msg="Everyone should die")
        pass

    def test_default_death_prob_zero(self):
        """
        Infect lots of people with cfr zero and short time to die
        duration. Verify that no one dies.
        Depends on default_cfr_one
        """
        total_agents = 500
        self.set_everyone_is_going_to_die(num_agents=total_agents)
        prob_dict = {
            DProgKeys.ProbabilityKeys.RelativeProbKeys.crt_to_death_probability: 0.0
        }
        self.set_simulation_prognosis_probability(prob_dict)
        self.run_sim()
        deaths_at_timestep_channel = self.get_full_result_channel(
            ResKeys.deaths_daily
        )
        deaths_cumulative_channel = self.get_full_result_channel(
            ResKeys.deaths_cumulative
        )
        death_channels = [
            deaths_at_timestep_channel,
            deaths_cumulative_channel
        ]
        for c in death_channels:
            for t in range(len(c)):
                self.assertEqual(c[t], 0,
                                 msg=f"There should be no deaths"
                                     f" with critical to death probability 0.0. Channel {c} had"
                                     f" bad data at t: {t}")
                pass
            pass
        cumulative_recoveries = self.get_day_final_channel_value(
            ResKeys.recovered_cumulative
        )
        self.assertGreaterEqual(cumulative_recoveries, 200,
                                msg="Should be lots of recoveries")
        pass

    def test_default_death_prob_scaling(self):
        """
        Infect lots of people with cfr zero and short time to die
        duration. Verify that no one dies.
        Depends on default_cfr_one
        """
        total_agents = 500
        self.set_everyone_is_going_to_die(num_agents=total_agents)
        end_sample_size = 10 # last few days most interesting
        death_probs = [0.01, 0.05, 0.10, 0.15]
        # old_ratio_sum = 0
        old_cumulative_deaths = 0
        for death_prob in death_probs:
            prob_dict = {
                DProgKeys.ProbabilityKeys.RelativeProbKeys.crt_to_death_probability: death_prob
            }
            self.set_simulation_prognosis_probability(prob_dict)
            self.run_sim()
            deaths_at_timestep_channel = self.get_full_result_channel(
                ResKeys.deaths_daily
            )
            recoveries_at_timestep_channel = self.get_full_result_channel(
                ResKeys.recovered_at_timestep
            )
            # new_ratio = []
            # for x in range(len(deaths_at_timestep_channel) - end_sample_size, len(deaths_at_timestep_channel)):
            #     new_ratio.append(deaths_at_timestep_channel[x]/
            #                      recoveries_at_timestep_channel[x])
            #     pass
            # self.assertGreater(sum(new_ratio), old_ratio_sum,
            #                    msg="As cfr increases, ratio should increase")
            cumulative_deaths = self.get_day_final_channel_value(
                ResKeys.deaths_cumulative
            )
            self.assertGreaterEqual(cumulative_deaths, old_cumulative_deaths,
                                    msg="Should be more deaths with higer ratio")
            old_cumulative_deaths = cumulative_deaths
            # old_ratio_sum = sum(new_ratio)
        pass

    @unittest.skip("P3")
    def test_cfr_by_age(self):
        pass

    # TODO: Define these as per parameter definitions


'''
Defines the Person class and functions associated with making people.
'''

#%% Imports
import numpy as np
import sciris as sc
from collections import defaultdict
from . import utils as cvu
from . import defaults as cvd
from . import base as cvb
from . import plotting as cvplt


__all__ = ['People']

class People(cvb.BasePeople):
    '''
    A class to perform all the operations on the people. This class is usually
    not invoked directly, but instead is created automatically by the sim. The
    only required input argument is the population size, but typically the full
    parameters dictionary will get passed instead since it will be needed before
    the People object is initialized.

    Args:
        pars (dict): the sim parameters, e.g. sim.pars -- alternatively, if a number, interpreted as pop_size
        strict (bool): whether or not to only create keys that are already in self.meta.person; otherwise, let any key be set
        kwargs (dict): the actual data, e.g. from a popdict, being specified

    ::Examples::

        ppl1 = cv.People(2000)

        sim = cv.Sim()
        ppl2 = cv.People(sim.pars)
    '''

    def __init__(self, pars, strict=True, **kwargs):

        # Handle pars and population size
        if sc.isnumber(pars): # Interpret as a population size
            pars = {'pop_size':pars} # Ensure it's a dictionary
        self.pars     = pars # Equivalent to self.set_pars(pars)
        self.pop_size = int(pars['pop_size'])

        # Other initialization
        self.t = 0 # Keep current simulation time
        self._lock = False # Prevent further modification of keys
        self.meta = cvd.PeopleMeta() # Store list of keys and dtypes
        self.contacts = None
        self.init_contacts() # Initialize the contacts
        self.infection_log = [] # Record of infections - keys for ['source','target','date','layer']

        # Set person properties -- all floats except for UID
        for key in self.meta.person:
            if key == 'uid':
                self[key] = np.arange(self.pop_size, dtype=cvd.default_int)
            else:
                self[key] = np.full(self.pop_size, np.nan, dtype=cvd.default_float)

        # Set health states -- only susceptible is true by default -- booleans
        for key in self.meta.states:
            if key == 'susceptible':
                self[key] = np.full(self.pop_size, True, dtype=bool)
            else:
                self[key] = np.full(self.pop_size, False, dtype=bool)

        # Set dates and durations -- both floats
        for key in self.meta.dates + self.meta.durs:
            self[key] = np.full(self.pop_size, np.nan, dtype=cvd.default_float)

        # Store the dtypes used in a flat dict
        self._dtypes = {key:self[key].dtype for key in self.keys()} # Assign all to float by default
        self._lock = True # Stop further keys from being set (does not affect attributes)

        # Store flows to be computed during simulation
        self.flows = {key:0 for key in cvd.new_result_flows}

        # Although we have called init(), we still need to call initialize()
        self.initialized = False

        # Handle contacts, if supplied (note: they usually are)
        if 'contacts' in kwargs:
            self.add_contacts(kwargs.pop('contacts'))

        # Handle all other values, e.g. age
        for key,value in kwargs.items():
            if strict:
                self.set(key, value)
            else:
                self[key] = value

        self._pending_quarantine = defaultdict(list)  # Internal cache to record people that need to be quarantined on each timestep {t:(person, quarantine_end_day)}
        return


    def initialize(self):
        ''' Perform initializations '''
        self.set_prognoses()
        self.validate()
        self.initialized = True
        return


    def set_prognoses(self):
        '''
        Set the prognoses for each person based on age during initialization. Need
        to reset the seed because viral loads are drawn stochastically.
        '''

        pars = self.pars # Shorten
        if 'prognoses' not in pars:
            errormsg = 'This people object does not have the required parameters ("prognoses"). Create a sim (or parameters), then do e.g. people.set_pars(sim.pars).'
            raise sc.KeyNotFoundError(errormsg)

        def find_cutoff(age_cutoffs, age):
            '''
            Find which age bin each person belongs to -- e.g. with standard
            age bins 0, 10, 20, etc., ages [5, 12, 4, 58] would be mapped to
            indices [0, 1, 0, 5]. Age bins are not guaranteed to be uniform
            width, which is why this can't be done as an array operation.
            '''
            return np.nonzero(age_cutoffs <= age)[0][-1]  # Index of the age bin to use

        cvu.set_seed(pars['rand_seed'])

        progs = pars['prognoses'] # Shorten the name
        inds = np.fromiter((find_cutoff(progs['age_cutoffs'], this_age) for this_age in self.age), dtype=cvd.default_int, count=len(self)) # Convert ages to indices
        self.symp_prob[:]   = progs['symp_probs'][inds] # Probability of developing symptoms
        self.severe_prob[:] = progs['severe_probs'][inds]*progs['comorbidities'][inds] # Severe disease probability is modified by comorbidities
        self.crit_prob[:]   = progs['crit_probs'][inds] # Probability of developing critical disease
        self.death_prob[:]  = progs['death_probs'][inds] # Probability of death
        self.rel_sus[:]     = progs['sus_ORs'][inds] # Default susceptibilities
        self.rel_trans[:]   = progs['trans_ORs'][inds]*cvu.sample(**self.pars['beta_dist'], size=len(inds)) # Default transmissibilities, with viral load drawn from a distribution

        return


    def update_states_pre(self, t):
        ''' Perform all state updates at the current timestep '''

        # Initialize
        self.t = t
        self.is_exp = self.true('exposed') # For storing the interim values since used in every subsequent calculation

        # Perform updates
        self.flows  = {key:0 for key in cvd.new_result_flows}
        self.flows['new_infectious']  += self.check_infectious() # For people who are exposed and not infectious, check if they begin being infectious
        self.flows['new_symptomatic'] += self.check_symptomatic()
        self.flows['new_severe']      += self.check_severe()
        self.flows['new_critical']    += self.check_critical()
        self.flows['new_deaths']      += self.check_death()
        self.flows['new_recoveries']  += self.check_recovery()

        return

    def update_states_post(self):
        ''' Perform post-timestep updates '''
        self.flows['new_diagnoses']   += self.check_diagnosed()
        self.flows['new_quarantined'] += self.check_quar()
        del self.is_exp  # Tidy up

        return


    def update_contacts(self):
        ''' Refresh dynamic contacts, e.g. community '''

        # Figure out if anything needs to be done -- e.g. {'h':False, 'c':True}
        dynam_keys = [lkey for lkey,is_dynam in self.pars['dynam_layer'].items() if is_dynam]

        # Loop over dynamic keys
        for lkey in dynam_keys:
            # Remove existing contacts
            self.contacts.pop(lkey)

            # Choose how many contacts to make
            pop_size   = len(self)
            n_contacts = self.pars['contacts'][lkey]
            n_new = int(n_contacts*pop_size/2) # Since these get looped over in both directions later

            # Create the contacts
            new_contacts = {} # Initialize
            new_contacts['p1']   = np.array(cvu.choose_r(max_n=pop_size, n=n_new), dtype=cvd.default_int) # Choose with replacement
            new_contacts['p2']   = np.array(cvu.choose_r(max_n=pop_size, n=n_new), dtype=cvd.default_int)
            new_contacts['beta'] = np.ones(n_new, dtype=cvd.default_float)

            # Add to contacts
            self.add_contacts(new_contacts, lkey=lkey)
            self.contacts[lkey].validate()

        return self.contacts


    #%% Methods for updating state

    def check_inds(self, current, date, filter_inds=None):
        ''' Return indices for which the current state is false and which meet the date criterion '''
        if filter_inds is None:
            not_current = cvu.false(current)
        else:
            not_current = cvu.ifalsei(current, filter_inds)
        has_date = cvu.idefinedi(date, not_current)
        inds     = cvu.itrue(self.t >= date[has_date], has_date)
        return inds


    def check_infectious(self):
        ''' Check if they become infectious '''
        inds = self.check_inds(self.infectious, self.date_infectious, filter_inds=self.is_exp)
        self.infectious[inds] = True
        return len(inds)


    def check_symptomatic(self):
        ''' Check for new progressions to symptomatic '''
        inds = self.check_inds(self.symptomatic, self.date_symptomatic, filter_inds=self.is_exp)
        self.symptomatic[inds] = True
        return len(inds)


    def check_severe(self):
        ''' Check for new progressions to severe '''
        inds = self.check_inds(self.severe, self.date_severe, filter_inds=self.is_exp)
        self.severe[inds] = True
        return len(inds)


    def check_critical(self):
        ''' Check for new progressions to critical '''
        inds = self.check_inds(self.critical, self.date_critical, filter_inds=self.is_exp)
        self.critical[inds] = True
        return len(inds)


    def check_recovery(self):
        ''' Check for recovery '''
        inds = self.check_inds(self.recovered, self.date_recovered, filter_inds=self.is_exp)
        self.exposed[inds]     = False
        self.infectious[inds]  = False
        self.symptomatic[inds] = False
        self.severe[inds]      = False
        self.critical[inds]    = False
        self.recovered[inds]   = True
        return len(inds)


    def check_death(self):
        ''' Check whether or not this person died on this timestep  '''
        inds = self.check_inds(self.dead, self.date_dead, filter_inds=self.is_exp)
        self.exposed[inds]     = False
        self.infectious[inds]  = False
        self.symptomatic[inds] = False
        self.severe[inds]      = False
        self.critical[inds]    = False
        self.recovered[inds]   = False
        self.dead[inds]        = True
        return len(inds)


    def check_diagnosed(self):
        '''
        Check for new diagnoses. Since most data are reported with diagnoses on
        the date of the test, this function reports counts not for the number of
        people who received a positive test result on a day, but rather, the number
        of people who were tested on that day who are schedule to be diagnosed in
        the future.
        '''

        # Handle people who tested today who will be diagnosed in future
        test_pos_inds = self.check_inds(self.diagnosed, self.date_pos_test, filter_inds=None) # Find people who will be diagnosed in future
        self.date_pos_test[test_pos_inds] = np.nan # Clear date of having will-be-positive test

        # Handle people who were actually diagnosed today
        diag_inds  = self.check_inds(self.diagnosed, self.date_diagnosed, filter_inds=None) # Find who was actually diagnosed on this timestep
        self.diagnosed[diag_inds]   = True # Set these people to be diagnosed
        quarantined = cvu.itruei(self.quarantined, diag_inds)
        self.date_end_quarantine[quarantined] = self.t # Set end quarantine date to match when the person left quarantine (and entered isolation)
        self.quarantined[diag_inds] = False # If you are diagnosed, you are isolated, not in quarantine

        return len(test_pos_inds)


    def check_quar(self):
        ''' Update quarantine state '''

        n_quarantined = 0 # Number of people entering quarantine
        for ind,end_day in self._pending_quarantine[self.t]:
            if self.quarantined[ind]:
                self.date_end_quarantine[ind] = max(self.date_end_quarantine[ind], end_day) # Extend quarantine if required
            elif not (self.dead[ind] or self.recovered[ind] or self.diagnosed[ind]): # Unclear whether recovered should be included here
                self.quarantined[ind] = True
                self.date_quarantined[ind] = self.t
                self.date_end_quarantine[ind] = end_day
                n_quarantined += 1

        # If someone on quarantine has reached the end of their quarantine, release them
        end_inds = self.check_inds(~self.quarantined, self.date_end_quarantine, filter_inds=None) # Note the double-negative here (~)
        self.quarantined[end_inds] = False # Release from quarantine

        return n_quarantined


    #%% Methods to make events occur (infection and diagnosis)

    def make_susceptible(self, inds):
        '''
        Make a set of people susceptible. This is used during dynamic resampling.
        '''
        for key in self.meta.states:
            if key == 'susceptible':
                self[key][inds] = True
            else:
                self[key][inds] = False

        for key in self.meta.dates + self.meta.durs:
            self[key][inds] = np.nan

        return


    def infect(self, inds, hosp_max=None, icu_max=None, source=None, layer=None):
        '''
        Infect people and determine their eventual outcomes.
            * Every infected person can infect other people, regardless of whether they develop symptoms
            * Infected people that develop symptoms are disaggregated into mild vs. severe (=requires hospitalization) vs. critical (=requires ICU)
            * Every asymptomatic, mildly symptomatic, and severely symptomatic person recovers
            * Critical cases either recover or die

        Args:
            inds     (array): array of people to infect
            hosp_max (bool):  whether or not there is an acute bed available for this person
            icu_max  (bool):  whether or not there is an ICU bed available for this person
            source   (array): source indices of the people who transmitted this infection (None if an importation or seed infection)
            layer    (str):   contact layer this infection was transmitted on

        Returns:
            count (int): number of people infected
        '''

        # Remove duplicates
        unique = np.unique(inds, return_index=True)[1]
        inds = inds[unique]
        if source is not None:
            source = source[unique]

        # Keep only susceptibles
        keep = self.susceptible[inds] # Unique indices in inds and source that are also susceptible
        inds = inds[keep]
        if source is not None:
            source = source[keep]

        n_infections = len(inds)
        durpars      = self.pars['dur']

        # Set states
        self.susceptible[inds]   = False
        self.exposed[inds]       = True
        self.date_exposed[inds]  = self.t
        self.flows['new_infections'] += len(inds)

        # Record transmissions
        for i, target in enumerate(inds):
            self.infection_log.append(dict(source=source[i] if source is not None else None, target=target, date=self.t, layer=layer))

        # Calculate how long before this person can infect other people
        self.dur_exp2inf[inds] = cvu.sample(**durpars['exp2inf'], size=n_infections)
        self.date_infectious[inds] = self.dur_exp2inf[inds] + self.t

        # Use prognosis probabilities to determine what happens to them
        symp_probs = self.pars['rel_symp_prob']*self.symp_prob[inds] # Calculate their actual probability of being symptomatic
        is_symp = cvu.binomial_arr(symp_probs) # Determine if they develop symptoms
        symp_inds = inds[is_symp]
        asymp_inds = inds[~is_symp] # Asymptomatic

        # CASE 1: Asymptomatic: may infect others, but have no symptoms and do not die
        dur_asym2rec = cvu.sample(**durpars['asym2rec'], size=len(asymp_inds))
        self.date_recovered[asymp_inds] = self.date_infectious[asymp_inds] + dur_asym2rec  # Date they recover
        self.dur_disease[asymp_inds] = self.dur_exp2inf[asymp_inds] + dur_asym2rec  # Store how long this person had COVID-19

        # CASE 2: Symptomatic: can either be mild, severe, or critical
        n_symp_inds = len(symp_inds)
        self.dur_inf2sym[symp_inds] = cvu.sample(**durpars['inf2sym'], size=n_symp_inds) # Store how long this person took to develop symptoms
        self.date_symptomatic[symp_inds] = self.date_infectious[symp_inds] + self.dur_inf2sym[symp_inds] # Date they become symptomatic
        sev_probs = self.pars['rel_severe_prob'] * self.severe_prob[symp_inds] # Probability of these people being severe
        is_sev = cvu.binomial_arr(sev_probs) # See if they're a severe or mild case
        sev_inds = symp_inds[is_sev]
        mild_inds = symp_inds[~is_sev] # Not severe

        # CASE 2.1: Mild symptoms, no hospitalization required and no probaility of death
        dur_mild2rec = cvu.sample(**durpars['mild2rec'], size=len(mild_inds))
        self.date_recovered[mild_inds] = self.date_symptomatic[mild_inds] + dur_mild2rec  # Date they recover
        self.dur_disease[mild_inds] = self.dur_exp2inf[mild_inds] + self.dur_inf2sym[mild_inds] + dur_mild2rec  # Store how long this person had COVID-19

        # CASE 2.2: Severe cases: hospitalization required, may become critical
        self.dur_sym2sev[sev_inds] = cvu.sample(**durpars['sym2sev'], size=len(sev_inds)) # Store how long this person took to develop severe symptoms
        self.date_severe[sev_inds] = self.date_symptomatic[sev_inds] + self.dur_sym2sev[sev_inds]  # Date symptoms become severe
        crit_probs = self.pars['rel_crit_prob'] * self.crit_prob[sev_inds] * (self.pars['no_hosp_factor'] if hosp_max else 1.)# Probability of these people becoming critical - higher if no beds available
        is_crit = cvu.binomial_arr(crit_probs)  # See if they're a critical case
        crit_inds = sev_inds[is_crit]
        non_crit_inds = sev_inds[~is_crit]

        # CASE 2.2.1 Not critical - they will recover
        dur_sev2rec = cvu.sample(**durpars['sev2rec'], size=len(non_crit_inds))
        self.date_recovered[non_crit_inds] = self.date_severe[non_crit_inds] + dur_sev2rec  # Date they recover
        self.dur_disease[non_crit_inds] = self.dur_exp2inf[non_crit_inds] + self.dur_inf2sym[non_crit_inds] + self.dur_sym2sev[non_crit_inds] + dur_sev2rec  # Store how long this person had COVID-19

        # CASE 2.2.2: Critical cases: ICU required, may die
        self.dur_sev2crit[crit_inds] = cvu.sample(**durpars['sev2crit'], size=len(crit_inds))
        self.date_critical[crit_inds] = self.date_severe[crit_inds] + self.dur_sev2crit[crit_inds]  # Date they become critical
        death_probs = self.pars['rel_death_prob'] * self.death_prob[crit_inds] * (self.pars['no_icu_factor'] if icu_max else 1.) # Probability they'll die
        is_dead = cvu.binomial_arr(death_probs)  # Death outcome
        dead_inds = crit_inds[is_dead]
        alive_inds = crit_inds[~is_dead]

        # CASE 2.2.2.1: Did not die
        dur_crit2rec = cvu.sample(**durpars['crit2rec'], size=len(alive_inds))
        self.date_recovered[alive_inds] = self.date_critical[alive_inds] + dur_crit2rec # Date they recover
        self.dur_disease[alive_inds] = self.dur_exp2inf[alive_inds] + self.dur_inf2sym[alive_inds] + self.dur_sym2sev[alive_inds] + self.dur_sev2crit[alive_inds] + dur_crit2rec  # Store how long this person had COVID-19

        # CASE 2.2.2.2: Did die
        dur_crit2die = cvu.sample(**durpars['crit2die'], size=len(dead_inds))
        self.date_dead[dead_inds] = self.date_critical[dead_inds] + dur_crit2die # Date of death
        self.dur_disease[dead_inds] = self.dur_exp2inf[dead_inds] + self.dur_inf2sym[dead_inds] + self.dur_sym2sev[dead_inds] + self.dur_sev2crit[dead_inds] + dur_crit2die   # Store how long this person had COVID-19

        return n_infections # For incrementing counters


    def test(self, inds, test_sensitivity=1.0, loss_prob=0.0, test_delay=0):
        '''
        Method to test people. Typically not to be called by the user directly;
        see the test_num() and test_prob() interventions.

        Args:
            inds: indices of who to test
            test_sensitivity (float): probability of a true positive
            loss_prob (float): probability of loss to follow-up
            test_delay (int): number of days before test results are ready
        '''

        inds = np.unique(inds)
        self.tested[inds] = True
        self.date_tested[inds] = self.t # Only keep the last time they tested

        is_infectious = cvu.itruei(self.infectious, inds)
        pos_test      = cvu.n_binomial(test_sensitivity, len(is_infectious))
        is_inf_pos    = is_infectious[pos_test]

        not_diagnosed = is_inf_pos[np.isnan(self.date_diagnosed[is_inf_pos])]
        not_lost      = cvu.n_binomial(1.0-loss_prob, len(not_diagnosed))
        final_inds    = not_diagnosed[not_lost]

        # Store the date the person will be diagnosed, as well as the date they took the test which will come back positive
        self.date_diagnosed[final_inds] = self.t + test_delay
        self.date_pos_test[final_inds] = self.t

        return


    def schedule_quarantine(self, inds, start_date=None, period=None):
        '''
        Schedule a quarantine. Typically not called by the user directly except
        via a custom intervention; see the contact_tracing() intervention instead.

        This function will create a request to quarantine a person on the start_date for
        a period of time. Whether they are on an existing quarantine that gets extended, or
        whether they are no longer eligible for quarantine, will be checked when the start_date
        is reached.

        Args:
            inds (int): indices of who to quarantine, specified by check_quar()
            start_date (int): day to begin quarantine (defaults to the current day, `sim.t`)
            period (int): quarantine duration (defaults to `pars['quar_period']`)
        '''

        start_date = self.t if start_date is None else int(start_date)
        period = self.pars['quar_period'] if period is None else int(period)
        for ind in inds:
            self._pending_quarantine[start_date].append((ind, start_date + period))
        return


    #%% Analysis methods

    def plot(self, *args, **kwargs):
        '''
        Plot statistics of the population -- age distribution, numbers of contacts,
        and overall weight of contacts (number of contacts multiplied by beta per
        layer).

        Args:
            bins (arr): age bins to use (default, 0-100 in one-year bins)
            width (float): bar width
            font_size (float): size of font
            alpha (float): transparency of the plots
            fig_args (dict): passed to pl.figure()
            axis_args (dict): passed to pl.subplots_adjust()
            plot_args (dict): passed to pl.plot()
        '''
        fig = cvplt.plot_people(people=self, *args, **kwargs)
        return fig


    def story(self, uid, *args):
        '''
        Print out a short history of events in the life of the specified individual.

        Args:
            uid (int/list): the person or people whose story is being regaled
            args (list): these people will tell their stories too

        **Example**::

            sim = cv.Sim(pop_type='hybrid', verbose=0)
            sim.run()
            sim.people.story(12)
            sim.people.story(795)
        '''

        def label_lkey(lkey):
            ''' Friendly name for common layer keys '''
            if lkey.lower() == 'a':
                llabel = 'default contact'
            if lkey.lower() == 'h':
                llabel = 'household'
            elif lkey.lower() == 's':
                llabel = 'school'
            elif lkey.lower() == 'w':
                llabel = 'workplace'
            elif lkey.lower() == 'c':
                llabel = 'community'
            else:
                llabel = f'"{lkey}"'
            return llabel

        uids = sc.promotetolist(uid)
        uids.extend(args)

        for uid in uids:

            p = self[uid]
            sex = 'female' if p.sex == 0 else 'male'

            intro = f'\nThis is the story of {uid}, a {p.age:.0f} year old {sex}'

            if not p.susceptible:
                if np.isnan(p.date_symptomatic):
                    print(f'{intro}, who had asymptomatic COVID.')
                else:
                    print(f'{intro}, who had symptomatic COVID.')
            else:
                print(f'{intro}, who did not contract COVID.')

            total_contacts = 0
            no_contacts = []
            for lkey in p.contacts.keys():
                llabel = label_lkey(lkey)
                n_contacts = len(p.contacts[lkey])
                total_contacts += n_contacts
                if n_contacts:
                    print(f'{uid} is connected to {n_contacts} people in the {llabel} layer')
                else:
                    no_contacts.append(llabel)
            if len(no_contacts):
                nc_string = ', '.join(no_contacts)
                print(f'{uid} has no contacts in the {nc_string} layer(s)')
            print(f'{uid} has {total_contacts} contacts in total')

            events = []

            dates = {
            'date_critical'       : 'became critically ill and needed ICU care',
            'date_dead'           : 'died ☹',
            'date_diagnosed'      : 'was diagnosed with COVID',
            'date_end_quarantine' : 'ended quarantine',
            'date_infectious'     : 'became infectious',
            'date_known_contact'  : 'was notified they may have been exposed to COVID',
            'date_pos_test'       : 'recieved their positive test result',
            'date_quarantined'    : 'entered quarantine',
            'date_recovered'      : 'recovered',
            'date_severe'         : 'developed severe symptoms and needed hospitalization',
            'date_symptomatic'    : 'became symptomatic',
            'date_tested'         : 'was tested for COVID',
            }

            for attribute, message in dates.items():
                date = getattr(p,attribute)
                if not np.isnan(date):
                    events.append((date, message))

            for infection in self.infection_log:
                lkey = infection['layer']
                llabel = label_lkey(lkey)
                if infection['target'] == uid:
                    if lkey:
                        events.append((infection['date'], f'was infected with COVID by {infection["source"]} via the {llabel} layer'))
                    else:
                        events.append((infection['date'], f'was infected with COVID as a seed infection'))

                if infection['source'] == uid:
                    x = len([a for a in self.infection_log if a['source'] == infection['target']])
                    events.append((infection['date'],f'gave COVID to {infection["target"]} via the {llabel} layer ({x} secondary infections)'))

            if len(events):
                for day, event in sorted(events, key=lambda x: x[0]):
                    print(f'On day {day:.0f}, {uid} {event}')
            else:
                print(f'Nothing happened to {uid} during the simulation.')
        return


import sciris as sc
import covasim as cv

cv.check_version('0.27.12')

pars = sc.objdict(
    pop_size     = 100e3, # Population size
    pop_infected = 1,     # Number of initial infections
    n_days       = 60,   # Number of days to simulate
    rand_seed    = 1,     # Random seed
    pop_type     = 'random',
    verbose      = 0,
)

sim = cv.Sim(pars=pars)

sc.tic()
sim.initialize()
sc.toc()

#%% Version 0.27.11
'''
>>> import sciris as sc
>>> import covasim as cv
Covasim 0.27.11 (2020-04-17) — © 2020 by IDM
>>> sim = cv.Sim(pop_size=1e6)
>>> sc.tic(); sim.initialize(); sc.toc()
1587187959.2373662
Initializing sim with 1e+06 people for 60 days
Elapsed time: 230 s
'''

#%% Version 0.27.12
'''
>>> import sciris as sc
>>> import covasim as cv
Covasim 0.27.12 (2020-04-17) — © 2020 by IDM
>>> sim = cv.Sim(pop_size=1e6)
>>> sc.tic(); sim.initialize(); sc.toc()
1587191076.600737
Initializing sim with 1e+06 people for 60 days
Elapsed time: 12.7 s
'''




'''
This script creates a single file containing all the scraped 
data from the Covid Data Project Data Scraper.
https://covidtracking.com
'''

from cova_epi_scraper import Scraper
import datetime as dt

def covid_tracking_date_to_date(d):
    return dt.date((d // 10000), ((d % 1000) // 100), (d % 1000) % 100)

class CovidTrackingProjectScraper(Scraper):
    def create_date(self):
        self.df['date'] = self.df.date.apply(covid_tracking_date_to_date)


class CovidUSTrackingProjectScraper(CovidTrackingProjectScraper):
    def create_key(self):
        self.df['key'] = 'US'


## Set up parameters 
pars_state = dict()
pars_state['title']         = "Covid Tracking Project Scraper for US states"
pars_state['load_path']     = "https://covidtracking.com/api/v1/states/daily.csv"
pars_state['output_folder'] = "epi_data/covid-tracking"

pars_state['renames'] = dict()
pars_state['renames']['state']                    = "key"
pars_state['renames']['positiveIncrease']         = "new_diagnoses"
pars_state['renames']['negativeIncrease']         = "new_negatives"
pars_state['renames']['totalTestResultsIncrease'] = "new_tests"
pars_state['renames']['hospitalizedIncrease']     = "new_hospitalized"
pars_state['renames']['deathIncrease']            = "new_deaths"
pars_state['renames']['inIcuCumulative']          = "cum_icu"
pars_state['renames']['hospitalizedCumulative']   = "cum_hospitalized"
pars_state['renames']['onVentilatorCumulative']   = "cum_on_ventilator"

pars_state['cumulative_fields'] = dict()
pars_state['cumulative_fields']['cum_icu']           = "new_icu"
pars_state['cumulative_fields']['cum_on_ventilator'] = "num_on_ventilator"


pars_state['fields_to_drop'] = [
    "hash",
    "dateChecked",
    "fips",
    "totalTestResults",
    "posNeg",
    "positive",
    "negative",
    "pending",
    "hospitalizedCurrently",
    "inIcuCurrently",
    "onVentilatorCurrently",
    "recovered",
    "hospitalized",
    "total"
]

# Set up US parameters
pars_us = dict()
pars_us['title']         = "Covid Tracking Project Scraper for US states"
pars_us['load_path']     = "https://covidtracking.com/api/v1/us/daily.csv"
pars_us['output_folder'] = "epi_data/covid-tracking"

pars_us['renames'] = dict()
pars_us['renames']['positiveIncrease']         = "new_diagnoses"
pars_us['renames']['negativeIncrease']         = "new_negatives"
pars_us['renames']['totalTestResultsIncrease'] = "new_tests"
pars_us['renames']['hospitalizedIncrease']     = "new_hospitalized"
pars_us['renames']['deathIncrease']            = "new_deaths"
pars_us['renames']['inIcuCumulative']          = "cum_icu"
pars_us['renames']['hospitalizedCumulative']   = "cum_hospitalized"
pars_us['renames']['onVentilatorCumulative']   = "cum_on_ventilator"

pars_us['cumulative_fields'] = dict()
pars_us['cumulative_fields']['cum_icu']           = "new_icu"
pars_us['cumulative_fields']['cum_on_ventilator'] = "new_on_ventilator"


pars_us['fields_to_drop'] = [
    "hash",
    "dateChecked",
    "fips",
    "totalTestResults",
    "posNeg",
    "positive",
    "negative",
    "pending",
    "hospitalizedCurrently",
    "inIcuCurrently",
    "onVentilatorCurrently",
    "recovered",
    "hospitalized",
    "states",
    "total"
]

# Scrape states
CovidTrackingProjectScraper(pars_state).scrape()
# Scrape US
CovidUSTrackingProjectScraper(pars_us).scrape()





import unittest
import covasim as cv

class TestDataLoaders(unittest.TestCase):
    def test_country_households(self):
        ch = cv.data.get_household_size()

        self.assertTrue(isinstance(ch, dict))
        self.assertGreater(len(ch), 5)

        self.assertTrue(ch['united states of america'] == ch['usa'])
        self.assertTrue(ch['republic of korea'] == ch['korea'])
        self.assertTrue(ch['republic of korea'] == ch['south korea'])

        self.assertTrue(2 <= ch['usa'] <= 5)
        self.assertTrue(1 <= ch['korea'] <= 3)
        self.assertTrue(1 <= ch['germany'] <= 3)
        self.assertTrue(5 <= ch['senegal'] <= 10)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '1.7.4'
__versiondate__ = '2020-10-02'
__license__ = f'Covasim {__version__} ({__versiondate__}) — © 2020 by IDM'
